import 'dart:async';
import 'dart:ui' as ui;

import 'package:flutter/material.dart';
import 'package:navigine_sdk/navigine_sdk.dart';

class SubLocationWidget extends StatefulWidget {
  final SubLocation subLocation;
  final Function(String)? onBeaconTapped;
  final VoidCallback? onMapTapped;
  final double x, y;
  final String? polyline;
  final List<BeaconDot> beaconDots;

  const SubLocationWidget({
    Key? key,
    required this.subLocation,
    this.x = 0,
    this.y = 0,
    this.onMapTapped,
    this.onBeaconTapped,
    this.beaconDots = const [],
    this.polyline,
  }) : super(key: key);
  @override
  _SubLocationWidgetState createState() => _SubLocationWidgetState();
}

class _SubLocationWidgetState extends State<SubLocationWidget> {
  late SubLocation subLocation;
  Size imageSize = Size(1, 1);
  late Offset currentPosition;
  late List<BeaconDot> beaconDots;

  @override
  initState() {
    super.initState();
    getImage();
    beaconDots = widget.beaconDots;
    subLocation = widget.subLocation;
    currentPosition = Offset(
      widget.x / subLocation.width,
      widget.y / subLocation.height,
    );
  }

  getImage() async {
    Completer<ui.Image> completer = Completer<ui.Image>();
    MemoryImage(widget.subLocation.image!)
        .resolve(ImageConfiguration())
        .addListener(ImageStreamListener(
            (ImageInfo info, bool _) => completer.complete(info.image)));

    ui.Image i = await completer.future;

    setState(() {
      imageSize = Size(i.width.toDouble(), i.height.toDouble());
    });
  }

  @override
  void didUpdateWidget(SubLocationWidget oldWidget) {
    super.didUpdateWidget(oldWidget);
    setState(() {
      beaconDots = widget.beaconDots;

      currentPosition = Offset(
        widget.x / subLocation.width,
        widget.y / subLocation.height,
      );
    });
  }

  Beacon? shouldInvokeTap(Offset tapped, BuildContext context) {
    var size = MediaQuery.of(context).size;
    var widgetHeight = size.width / imageSize.width * imageSize.height;
    for (var b in subLocation.beacons) {
      var x = b.x / subLocation.width * size.width;
      var y = (1 - b.y / subLocation.height) * widgetHeight;
      if ((tapped.dx - x).abs() < 20 && (tapped.dy - y).abs() < 20) {
        return b;
      }
    }
    return null;
  }

  @override
  Widget build(BuildContext context) {
    var size = MediaQuery.of(context).size;
    return Stack(
      children: <Widget>[
        SizedBox(
          width: size.width,
          height: size.width / imageSize.width * imageSize.height,
          child: GestureDetector(
            onTapDown: (detail) {
              RenderBox? box = context.findRenderObject() as RenderBox?;
              Offset? local = box?.globalToLocal(detail.globalPosition);
              if (local != null) {
                Beacon? beacon = shouldInvokeTap(local, context);
                if (beacon != null) {
                  widget.onBeaconTapped?.call(beacon.uuid);
                } else {
                  widget.onMapTapped?.call();
                }
              }
            },
            child: Stack(
              clipBehavior: Clip.antiAlias,
              children: <Widget>[
                Image.memory(
                  subLocation.image!,
                  width: imageSize.width,
                ),
                CustomPaint(
                  size: Size(
                    size.width,
                    size.width / imageSize.width * imageSize.height,
                  ),
                  painter: RoutePainter(
                    imageSize: Size(subLocation.width, subLocation.height),
                    polyline: widget.polyline,
                    beacons: beaconDots,
                    position: currentPosition,
                  ),
                ),
              ],
            ),
          ),
        ),
        if (currentPosition.dx == 0 && currentPosition.dy == 0)
          Center(child: CircularProgressIndicator(strokeWidth: 2)),
      ],
    );
  }
}

class RoutePainter extends CustomPainter {
  // This is the hash of the polyline
  String? polyline;
  // This contains x,y for the users current location
  Offset? position;
  // THis is the size of the image, I am paintiing
  Size imageSize;
  // This is the locations of beacons that are on this map
  List<BeaconDot> beacons;

  // This is the painting style for beacon dot
  var beaconPainter = Paint()..color = Colors.red;

  // This is the painting style for live user dot
  var dotPainter = Paint()..color = Colors.blue;

  // This is the painting style for route line
  var routePainter = Paint()
    ..strokeWidth = 2
    ..style = PaintingStyle.stroke
    ..color = Colors.red;

  RoutePainter({
    this.beacons = const [],
    this.position,
    required this.imageSize,
    required this.polyline,
  });

  Offset getConvertedPoint(Offset point, Size size) {
    return Offset(point.dx * size.width, (1 - point.dy) * size.height);
  }

  @override
  void paint(Canvas canvas, Size size) {
    int i = 0;
    Path path = Path();

    // This function gives array or points from the hash created based on google maps
    List<Offset>? points;
    if (polyline != null) {
      points = decodePolylineToOffset(polyline!);

      // Drawing line using a path, first point is starting the path and else are drawing line to the path
      points.forEach((p) {
        // Navigine uses dy in opposite direction, thus this fuction does 1- dy and multiplies it by the width and height of the image.
        Offset converted = getConvertedPoint(p, size);
        if (i == 0)
          path.moveTo(converted.dx, converted.dy);
        else
          path.lineTo(converted.dx, converted.dy);
        i++;
      });
    }

    // Drawing circle on the map for each beacons

    beacons.forEach(
      (b) {
        Offset position = Offset(
          b.x / imageSize.width * size.width,
          (1 - b.y / imageSize.height) * size.height,
        );
        canvas.drawCircle(
          position,
          8,
          beaconPainter..color = b.color,
        );
        TextSpan span =
            TextSpan(style: TextStyle(color: Colors.blue), text: b.name);
        TextPainter tp = TextPainter(
            text: span,
            textAlign: TextAlign.left,
            textDirection: TextDirection.ltr);
        tp.layout();
        tp.paint(canvas, position.translate(-tp.width / 2, 10));
      },
    );

    // Drawing circle on the map for live user location
    if (position != null) {
      if (position!.dy != 0 && position!.dx != 0)
        canvas.drawCircle(
          Offset(position!.dx * size.width, (1 - position!.dy) * size.height),
          12,
          dotPainter,
        );
    }

    // Drawing the line after all the points, this will rise up the line ie z index is high
    canvas.drawPath(path, routePainter);
  }

  @override
  bool shouldRepaint(RoutePainter oldDelegate) {
    return true;
  }

  List<ui.Offset> decodePolylineToOffset(String polyline) {
    return [];
  }
}

class BeaconDot {
  final String name;
  final double x, y;
  final Color color;

  BeaconDot({
    required this.name,
    required this.x,
    required this.y,
    required this.color,
  });
}
